import inspect
import uuid
from copy import deepcopy
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Union,
    cast,
)

from pydantic import Field
from pydantic.dataclasses import dataclass
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from literalai.client import BaseLiteralClient
    from literalai.event_processor import EventProcessor

from literalai.context import active_steps_var, active_thread_var, active_root_run_var
from literalai.helper import utc_now
from literalai.my_types import Environment, Utils
from literalai.observability.generation import (
    BaseGeneration,
    ChatGeneration,
    CompletionGeneration,
)

TrueStepType = Literal[
    "run", "tool", "llm", "embedding", "retrieval", "rerank", "undefined"
]

MessageStepType = Literal["user_message", "assistant_message", "system_message"]

StepType = Union[TrueStepType, MessageStepType]


ScoreType = Literal["HUMAN", "AI"]


class ScoreDict(TypedDict, total=False):
    id: Optional[str]
    name: str
    type: ScoreType
    value: float
    stepId: Optional[str]
    datasetExperimentItemId: Optional[str]
    comment: Optional[str]
    tags: Optional[List[str]]


class AttachmentDict(TypedDict, total=False):
    id: Optional[str]
    stepId: Optional[str]
    threadId: Optional[str]
    metadata: Optional[Dict]
    mime: Optional[str]
    name: Optional[str]
    objectKey: Optional[str]
    url: Optional[str]


@dataclass(repr=False)
class Score(Utils):
    name: str
    type: ScoreType
    value: float
    step_id: Optional[str]
    dataset_experiment_item_id: Optional[str]
    comment: Optional[str]
    tags: Optional[List[str]]
    id: Optional[str] = Field(default_factory=lambda: str(uuid.uuid4()))

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "value": self.value,
            "stepId": self.step_id,
            "datasetExperimentItemId": self.dataset_experiment_item_id,
            "comment": self.comment,
            "tags": self.tags,
        }

    @classmethod
    def from_dict(cls, score_dict: ScoreDict) -> "Score":
        id = score_dict.get("id", "")
        name = score_dict.get("name", "")
        type = score_dict.get("type", "HUMAN")
        value = score_dict.get("value", 0.0)
        step_id = score_dict.get("stepId", "")
        dataset_experiment_item_id = score_dict.get("datasetExperimentItemId", "")
        comment = score_dict.get("comment", "")
        tags = score_dict.get("tags", [])

        score = cls(
            id=id,
            name=name,
            type=type,
            value=value,
            step_id=step_id,
            dataset_experiment_item_id=dataset_experiment_item_id,
            comment=comment,
            tags=tags,
        )

        return score


@dataclass(repr=False)
class Attachment(Utils):
    step_id: Optional[str] = None
    thread_id: Optional[str] = None
    id: Optional[str] = Field(default_factory=lambda: str(uuid.uuid4()))
    metadata: Optional[Dict] = Field(default_factory=lambda: {})
    mime: Optional[str] = None
    name: Optional[str] = None
    object_key: Optional[str] = None
    url: Optional[str] = None

    def to_dict(self):
        return {
            "id": self.id,
            "metadata": self.metadata,
            "mime": self.mime,
            "name": self.name,
            "objectKey": self.object_key,
            "url": self.url,
        }

    @classmethod
    def from_dict(cls, attachment_dict: AttachmentDict) -> "Attachment":
        id = attachment_dict.get("id", "")
        thread_id = attachment_dict.get("threadId", None)
        step_id = attachment_dict.get("stepId", None)
        metadata = attachment_dict.get("metadata", {})
        mime = attachment_dict.get("mime", "")
        name = attachment_dict.get("name", "")
        object_key = attachment_dict.get("objectKey", "")
        url = attachment_dict.get("url", "")

        attachment = cls(
            id=id,
            thread_id=thread_id,
            mime=mime,
            name=name,
            object_key=object_key,
            url=url,
            step_id=step_id,
            metadata=metadata,
        )

        return attachment


class StepDict(TypedDict, total=False):
    id: Optional[str]
    name: Optional[str]
    type: Optional[StepType]
    environment: Optional[Environment]
    threadId: Optional[str]
    error: Optional[str]
    input: Optional[Dict]
    output: Optional[Dict]
    metadata: Optional[Dict]
    tags: Optional[List[str]]
    parentId: Optional[str]
    createdAt: Optional[str]
    startTime: Optional[str]
    endTime: Optional[str]
    generation: Optional[Dict]
    scores: Optional[List[ScoreDict]]
    attachments: Optional[List[AttachmentDict]]
    rootRunId: Optional[str]


class Step(Utils):
    id: str
    name: Optional[str] = ""
    type: Optional[StepType] = None
    metadata: Optional[Dict] = None
    parent_id: Optional[str] = None
    start_time: Optional[str] = None
    end_time: Optional[str] = None
    created_at: Optional[str] = None
    error: Optional[str] = None
    input: Optional[Dict[str, Any]] = None
    output: Optional[Dict[str, Any]] = None
    tags: Optional[List[str]] = None
    thread_id: Optional[str] = None
    environment: Optional[Environment] = None
    root_run_id: Optional[str] = None

    generation: Optional[Union[ChatGeneration, CompletionGeneration]] = None
    scores: Optional[List[Score]] = []
    attachments: List[Attachment] = []

    def __init__(
        self,
        name: str = "",
        type: Optional[StepType] = None,
        id: Optional[str] = None,
        thread_id: Optional[str] = None,
        parent_id: Optional[str] = None,
        processor: Optional["EventProcessor"] = None,
        tags: Optional[List[str]] = None,
        root_run_id: Optional[str] = None,
    ):
        from time import sleep

        sleep(0.001)
        self.id = id or str(uuid.uuid4())
        self.start_time = utc_now()
        self.name = name
        self.type = type

        self.processor = processor

        # priority for thread_id: thread_id > parent_step.thread_id > active_thread
        self.thread_id = thread_id

        # priority for root_run_id: root_run_id > parent_step.root_run_id > active_root_run
        self.root_run_id = root_run_id

        # priority for parent_id: parent_id > parent_step.id
        self.parent_id = parent_id

        self.tags = tags

    def start(self):
        active_steps = active_steps_var.get()
        if len(active_steps) > 0:
            parent_step = active_steps[-1]
            if not self.parent_id:
                self.parent_id = parent_step.id
            if not self.thread_id:
                self.thread_id = parent_step.thread_id
            if not self.root_run_id:
                self.root_run_id = parent_step.root_run_id

        if not self.thread_id:
            if active_thread := active_thread_var.get():
                self.thread_id = active_thread.id

        if not self.root_run_id:
            if active_root_run := active_root_run_var.get():
                self.root_run_id = active_root_run.id

        active_steps.append(self)
        active_steps_var.set(active_steps)

    def end(self):
        self.end_time = utc_now()

        # Update active steps
        active_steps = active_steps_var.get()

        # Check if step is active
        if self not in active_steps:
            raise Exception("Step must be started before ending.")

        # Remove step from active steps
        active_steps.remove(self)
        active_steps_var.set(active_steps)

        if self.processor is None:
            raise Exception(
                "Step must be stopped with a processor to allow finalization."
            )
        self.processor.add_event(self.to_dict())

    def to_dict(self):
        return {
            "id": self.id,
            "metadata": self.metadata,
            "parentId": self.parent_id,
            "startTime": self.start_time,
            "endTime": self.end_time,
            "type": self.type,
            "threadId": self.thread_id,
            "error": self.error,
            "input": self.input,
            "output": self.output,
            "generation": self.generation.to_dict() if self.generation else None,
            "name": self.name,
            "tags": self.tags,
            "scores": [score.to_dict() for score in self.scores],
            "attachments": [attachment.to_dict() for attachment in self.attachments],
            "rootRunId": self.root_run_id,
        }

    @classmethod
    def from_dict(cls, step_dict: StepDict) -> "Step":
        name = step_dict.get("name") or ""
        step_type = step_dict.get("type", cast(StepType, "undefined"))
        thread_id = step_dict.get("threadId")

        step = cls(name=name, type=step_type, thread_id=thread_id)

        step.id = step_dict.get("id") or ""
        step.input = step_dict.get("input", None)
        step.error = step_dict.get("error", None)
        step.output = step_dict.get("output", None)
        step.environment = step_dict.get("environment", None)
        step.metadata = step_dict.get("metadata", {})
        step.tags = step_dict.get("tags", [])
        step.parent_id = step_dict.get("parentId", None)
        step.start_time = step_dict.get("startTime", None)
        step.end_time = step_dict.get("endTime", None)
        step.created_at = step_dict.get("createdAt", None)

        if "generation" in step_dict and step_type == "llm":
            generation_dict = step_dict["generation"]
            if generation_dict:
                step.generation = BaseGeneration.from_dict(generation_dict)

        if "scores" in step_dict:
            scores = step_dict["scores"]
            if scores:
                step.scores = [Score.from_dict(score) for score in scores]

        if "attachments" in step_dict:
            attachments = step_dict["attachments"]
            if attachments:
                step.attachments = [
                    Attachment.from_dict(attachment) for attachment in attachments
                ]

        return step


class StepContextManager:
    def __init__(
        self,
        client: "BaseLiteralClient",
        name: str = "",
        type: TrueStepType = "undefined",
        id: Optional[str] = None,
        parent_id: Optional[str] = None,
        thread_id: Optional[str] = None,
        root_run_id: Optional[str] = None,
        **kwargs,
    ):
        self.client = client
        self.step_name = name
        self.step_type = type
        self.id = id
        self.parent_id = parent_id
        self.thread_id = thread_id
        self.root_run_id = root_run_id
        self.kwargs = kwargs

    def __call__(self, func):
        return step_decorator(
            self.client,
            func=func,
            name=self.step_name,
            ctx_manager=self,
        )

    async def __aenter__(self):
        self.step = self.client.start_step(
            name=self.step_name,
            type=self.step_type,
            id=self.id,
            parent_id=self.parent_id,
            thread_id=self.thread_id,
            root_run_id=self.root_run_id,
            **self.kwargs,
        )

        if active_root_run_var.get() is None and self.step_type == "run":
            active_root_run_var.set(self.step)

        return self.step

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.step.error = str(exc_val)
            await self.client.event_processor.aflush()

        if active_root_run_var.get():
            active_root_run_var.set(None)

        self.step.end()

    def __enter__(self) -> Step:
        self.step = self.client.start_step(
            name=self.step_name,
            type=self.step_type,
            id=self.id,
            parent_id=self.parent_id,
            thread_id=self.thread_id,
            root_run_id=self.root_run_id,
            **self.kwargs,
        )

        if active_root_run_var.get() is None and self.step_type == "run":
            active_root_run_var.set(self.step)

        return self.step

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.step.error = str(exc_val)
            self.client.event_processor.flush()

        if active_root_run_var.get():
            active_root_run_var.set(None)

        self.step.end()


def flatten_args_kwargs(func, *args, **kwargs):
    signature = inspect.signature(func)
    bound_arguments = signature.bind(*args, **kwargs)
    bound_arguments.apply_defaults()
    return {k: deepcopy(v) for k, v in bound_arguments.arguments.items()}


def step_decorator(
    client: "BaseLiteralClient",
    func: Callable,
    type: TrueStepType = "undefined",
    name: str = "",
    id: Optional[str] = None,
    parent_id: Optional[str] = None,
    thread_id: Optional[str] = None,
    root_run_id: Optional[str] = None,
    ctx_manager: Optional[StepContextManager] = None,
    **decorator_kwargs,
):
    if not name:
        name = func.__name__
    if not ctx_manager:
        ctx_manager = StepContextManager(
            client=client,
            type=type,
            name=name,
            id=id,
            parent_id=parent_id,
            thread_id=thread_id,
            root_run_id=root_run_id,
            **decorator_kwargs,
        )
    else:
        ctx_manager.step_name = name
    # Handle async decorator
    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            with ctx_manager as step:
                try:
                    step.input = flatten_args_kwargs(func, *args, **kwargs)
                except Exception:
                    pass
                result = await func(*args, **kwargs)
                try:
                    if step.output is None:
                        if isinstance(result, dict):
                            step.output = deepcopy(result)
                        else:
                            step.output = {"content": deepcopy(result)}
                except Exception:
                    pass
                return result

        return async_wrapper
    else:
        # Handle sync decorator
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            with ctx_manager as step:
                try:
                    step.input = flatten_args_kwargs(func, *args, **kwargs)
                except Exception:
                    pass
                result = func(*args, **kwargs)
                try:
                    if step.output is None:
                        if isinstance(result, dict):
                            step.output = deepcopy(result)
                        else:
                            step.output = {"content": deepcopy(result)}
                except Exception:
                    pass
                return result

        return sync_wrapper



import inspect
import os
from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional

from literalai.my_types import Environment

if TYPE_CHECKING:
    from literalai.client import BaseLiteralClient


class EnvContextManager:
    def __init__(self, client: "BaseLiteralClient", env: Environment = "prod"):
        self.client = client
        self.env = env
        self.original_env = os.environ.get("LITERAL_ENV", "")

    def __call__(self, func):
        return env_decorator(
            self.client,
            func=func,
            ctx_manager=self,
        )

    async def __aenter__(self):
        os.environ["LITERAL_ENV"] = self.env

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        os.environ = self.original_env

    def __enter__(self):
        os.environ["LITERAL_ENV"] = self.env

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.environ["LITERAL_ENV"] = self.original_env


def env_decorator(
    client: "BaseLiteralClient",
    func: Callable,
    env: Environment = "prod",
    ctx_manager: Optional[EnvContextManager] = None,
    **decorator_kwargs,
):
    if not ctx_manager:
        ctx_manager = EnvContextManager(
            client=client,
            env=env,
            **decorator_kwargs,
        )

    # Handle async decorator
    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            with ctx_manager:
                result = await func(*args, **kwargs)
                return result

        return async_wrapper
    else:
        # Handle sync decorator
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            with ctx_manager:
                return func(*args, **kwargs)

        return sync_wrapper



from typing import Any, Dict, List, Optional

from literalai.observability.filter import threads_filters, threads_order_by
from literalai.my_types import PaginatedResponse
from literalai.observability.step import StepType
from literalai.observability.thread import Thread

from literalai.api import gql


def get_threads_helper(
    first: Optional[int] = None,
    after: Optional[str] = None,
    before: Optional[str] = None,
    filters: Optional[threads_filters] = None,
    order_by: Optional[threads_order_by] = None,
    step_types_to_keep: Optional[List[StepType]] = None,
):
    variables: Dict[str, Any] = {}

    if first:
        variables["first"] = first
    if after:
        variables["after"] = after
    if before:
        variables["before"] = before
    if filters:
        variables["filters"] = filters
    if order_by:
        variables["orderBy"] = order_by
    if step_types_to_keep:
        variables["stepTypesToKeep"] = step_types_to_keep

    def process_response(response):
        processed_response = response["data"]["threads"]
        processed_response["data"] = [
            edge["node"] for edge in processed_response["edges"]
        ]
        return PaginatedResponse[Thread].from_dict(processed_response, Thread)

    description = "get threads"

    return gql.GET_THREADS, description, variables, process_response


def list_threads_helper(
    first: Optional[int] = None,
    after: Optional[str] = None,
    before: Optional[str] = None,
    filters: Optional[threads_filters] = None,
    order_by: Optional[threads_order_by] = None,
):
    variables: Dict[str, Any] = {}

    if first:
        variables["first"] = first
    if after:
        variables["after"] = after
    if before:
        variables["before"] = before
    if filters:
        variables["filters"] = filters
    if order_by:
        variables["orderBy"] = order_by

    def process_response(response):
        response_data = response["data"]["threads"]
        response_data["data"] = list(map(lambda x: x["node"], response_data["edges"]))
        return PaginatedResponse[Thread].from_dict(response_data, Thread)

    description = "get threads"

    return gql.LIST_THREADS, description, variables, process_response


def get_thread_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        thread = response["data"]["threadDetail"]
        return Thread.from_dict(thread) if thread else None

    description = "get thread"

    return gql.GET_THREAD, description, variables, process_response


def create_thread_helper(
    name: Optional[str] = None,
    metadata: Optional[Dict] = None,
    participant_id: Optional[str] = None,
    tags: Optional[List[str]] = None,
):
    variables = {
        "name": name,
        "metadata": metadata,
        "participantId": participant_id,
        "tags": tags,
    }

    def process_response(response):
        return Thread.from_dict(response["data"]["createThread"])

    description = "create thread"

    return gql.CREATE_THREAD, description, variables, process_response


def upsert_thread_helper(
    id: str,
    name: Optional[str] = None,
    metadata: Optional[Dict] = None,
    participant_id: Optional[str] = None,
    tags: Optional[List[str]] = None,
):
    variables = {
        "id": id,
        "name": name,
        "metadata": metadata,
        "participantId": participant_id,
        "tags": tags,
    }

    # remove None values to prevent the API from removing existing values
    variables = {k: v for k, v in variables.items() if v is not None}

    def process_response(response):
        return Thread.from_dict(response["data"]["upsertThread"])

    description = "upsert thread"

    return gql.UPSERT_THREAD, description, variables, process_response


def update_thread_helper(
    id: str,
    name: Optional[str] = None,
    metadata: Optional[Dict] = None,
    participant_id: Optional[str] = None,
    tags: Optional[List[str]] = None,
):
    variables = {
        "id": id,
        "name": name,
        "metadata": metadata,
        "participantId": participant_id,
        "tags": tags,
    }

    # remove None values to prevent the API from removing existing values
    variables = {k: v for k, v in variables.items() if v is not None}

    def process_response(response):
        return Thread.from_dict(response["data"]["updateThread"])

    description = "update thread"

    return gql.UPDATE_THREAD, description, variables, process_response


def delete_thread_helper(id: str):
    variables = {"thread_id": id}

    def process_response(response):
        deleted = bool(response["data"]["deleteThread"])
        return deleted

    description = "delete thread"

    # Assuming DELETE_THREAD is a placeholder for the actual GraphQL mutation
    return gql.DELETE_THREAD, description, variables, process_response



from typing import TYPE_CHECKING, Dict, Optional

from literalai.api import gql

from literalai.evaluation.dataset import Dataset, DatasetType
from literalai.evaluation.dataset_experiment import DatasetExperiment, DatasetExperimentItem
from literalai.evaluation.dataset_item import DatasetItem

if TYPE_CHECKING:
    from literalai.api import LiteralAPI



def create_dataset_helper(
    api: "LiteralAPI",
    name: str,
    description: Optional[str] = None,
    metadata: Optional[Dict] = None,
    type: DatasetType = "key_value",
):
    variables = {
        "name": name,
        "description": description,
        "metadata": metadata,
        "type": type,
    }

    def process_response(response):
        return Dataset.from_dict(api, response["data"]["createDataset"])

    description = "create dataset"

    return gql.CREATE_DATASET, description, variables, process_response


def get_dataset_helper(
    api: "LiteralAPI", id: Optional[str] = None, name: Optional[str] = None
):
    if not id and not name:
        raise ValueError("id or name must be provided")

    body = {}

    if id:
        body["id"] = id
    if name:
        body["name"] = name

    def process_response(response) -> Optional[Dataset]:
        dataset_dict = response.get("data")
        if dataset_dict is None:
            return None
        return Dataset.from_dict(api, dataset_dict)

    description = "get dataset"

    # Assuming there's a placeholder or a constant for the subpath
    subpath = "/export/dataset"

    return subpath, description, body, process_response


def update_dataset_helper(
    api: "LiteralAPI",
    id: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[Dict] = None,
):
    variables: Dict = {"id": id}
    if name is not None:
        variables["name"] = name
    if description is not None:
        variables["description"] = description
    if metadata is not None:
        variables["metadata"] = metadata

    def process_response(response):
        return Dataset.from_dict(api, response["data"]["updateDataset"])

    description = "update dataset"

    return gql.UPDATE_DATASET, description, variables, process_response


def delete_dataset_helper(api: "LiteralAPI", id: str):
    variables = {"id": id}

    def process_response(response):
        return Dataset.from_dict(api, response["data"]["deleteDataset"])

    description = "delete dataset"

    return gql.DELETE_DATASET, description, variables, process_response


def create_experiment_helper(
    api: "LiteralAPI",
    name: str,
    dataset_id: Optional[str] = None,
    prompt_id: Optional[str] = None,
    params: Optional[Dict] = None,
):
    variables = {
        "datasetId": dataset_id,
        "name": name,
        "promptId": prompt_id,
        "params": params,
    }

    def process_response(response):
        return DatasetExperiment.from_dict(
            api, response["data"]["createDatasetExperiment"]
        )

    description = "create dataset experiment"

    return gql.CREATE_EXPERIMENT, description, variables, process_response


def create_experiment_item_helper(
    dataset_experiment_id: str,
    experiment_run_id: Optional[str] = None,
    dataset_item_id: Optional[str] = None,
    input: Optional[Dict] = None,
    output: Optional[Dict] = None,
):
    variables = {
        "experimentRunId": experiment_run_id,
        "datasetExperimentId": dataset_experiment_id,
        "datasetItemId": dataset_item_id,
        "input": input,
        "output": output,
    }

    def process_response(response):
        return DatasetExperimentItem.from_dict(
            response["data"]["createDatasetExperimentItem"]
        )

    description = "create dataset experiment item"

    return gql.CREATE_EXPERIMENT_ITEM, description, variables, process_response


def create_dataset_item_helper(
    dataset_id: str,
    input: Dict,
    expected_output: Optional[Dict] = None,
    metadata: Optional[Dict] = None,
):
    variables = {
        "datasetId": dataset_id,
        "input": input,
        "expectedOutput": expected_output,
        "metadata": metadata,
    }

    def process_response(response):
        return DatasetItem.from_dict(response["data"]["createDatasetItem"])

    description = "create dataset item"

    return gql.CREATE_DATASET_ITEM, description, variables, process_response


def get_dataset_item_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        return DatasetItem.from_dict(response["data"]["datasetItem"])

    description = "get dataset item"

    return gql.GET_DATASET_ITEM, description, variables, process_response


def delete_dataset_item_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        return DatasetItem.from_dict(response["data"]["deleteDatasetItem"])

    description = "delete dataset item"

    return gql.DELETE_DATASET_ITEM, description, variables, process_response


def add_step_to_dataset_helper(
    dataset_id: str, step_id: str, metadata: Optional[Dict] = None
):
    variables = {
        "datasetId": dataset_id,
        "stepId": step_id,
        "metadata": metadata,
    }

    def process_response(response):
        return DatasetItem.from_dict(response["data"]["addStepToDataset"])

    description = "add step to dataset"

    return gql.ADD_STEP_TO_DATASET, description, variables, process_response


def add_generation_to_dataset_helper(
    dataset_id: str, generation_id: str, metadata: Optional[Dict] = None
):
    variables = {
        "datasetId": dataset_id,
        "generationId": generation_id,
        "metadata": metadata,
    }

    def process_response(response):
        return DatasetItem.from_dict(response["data"]["addGenerationToDataset"])

    description = "add generation to dataset"

    return gql.ADD_GENERATION_TO_DATASET, description, variables, process_response



import json
import uuid
from typing import Any, Dict, Generic, List, Literal, Optional, Protocol, TypeVar
from abc import abstractmethod

from typing_extensions import TypedDict

from pydantic.dataclasses import Field, dataclass

Environment = Literal["dev", "staging", "prod", "experiment"]


class Utils:
    def __str__(self):
        return json.dumps(self.to_dict(), sort_keys=True, indent=4)

    def __repr__(self):
        return json.dumps(self.to_dict(), sort_keys=True, indent=4)

    @abstractmethod
    def to_dict(self):
        pass


@dataclass(repr=False)
class PageInfo(Utils):
    has_next_page: bool
    start_cursor: Optional[str]
    end_cursor: Optional[str]

    def to_dict(self):
        return {
            "hasNextPage": self.has_next_page,
            "startCursor": self.start_cursor,
            "endCursor": self.end_cursor,
        }

    @classmethod
    def from_dict(cls, page_info_dict: Dict) -> "PageInfo":
        has_next_page = page_info_dict.get("hasNextPage", False)
        start_cursor = page_info_dict.get("startCursor", None)
        end_cursor = page_info_dict.get("endCursor", None)
        return cls(
            has_next_page=has_next_page, start_cursor=start_cursor, end_cursor=end_cursor
        )


T = TypeVar("T", covariant=True)


class HasFromDict(Protocol[T]):
    @classmethod
    def from_dict(cls, obj_dict: Any) -> T:
        raise NotImplementedError()


@dataclass(repr=False)
class PaginatedResponse(Generic[T], Utils):
    page_info: PageInfo
    data: List[T]
    total_count: Optional[int] = None

    def to_dict(self):
        return {
            "pageInfo": self.page_info.to_dict(),
            "totalCount": self.total_count,
            "data": [
                (d.to_dict() if hasattr(d, "to_dict") and callable(d.to_dict) else d)
                for d in self.data
            ],
        }

    @classmethod
    def from_dict(
        cls, paginated_response_dict: Dict, the_class: HasFromDict[T]
    ) -> "PaginatedResponse[T]":
        page_info = PageInfo.from_dict(paginated_response_dict.get("pageInfo", {}))
        data = [the_class.from_dict(d) for d in paginated_response_dict.get("data", [])]
        total_count = paginated_response_dict.get("totalCount", None)
        return cls(page_info=page_info, data=data, total_count=total_count)


class TextContent(TypedDict, total=False):
    type: Literal["text"]
    text: str


class ImageUrlContent(TypedDict, total=False):
    type: Literal["image_url"]
    image_url: Dict


class UserDict(TypedDict, total=False):
    id: Optional[str]
    metadata: Optional[Dict]
    identifier: Optional[str]
    createdAt: Optional[str]


@dataclass(repr=False)
class User(Utils):
    id: Optional[str] = None
    created_at: Optional[str] = None
    identifier: Optional[str] = Field(default_factory=lambda: str(uuid.uuid4()))
    metadata: Dict = Field(default_factory=lambda: {})

    def to_dict(self) -> UserDict:
        return {
            "id": self.id,
            "identifier": self.identifier,
            "metadata": self.metadata,
            "createdAt": self.created_at,
        }

    @classmethod
    def from_dict(cls, user_dict: Dict) -> "User":
        id_ = user_dict.get("id", "")
        identifier = user_dict.get("identifier", "")
        metadata = user_dict.get("metadata", {})
        created_at = user_dict.get("createdAt", "")

        user = cls(
            id=id_, identifier=identifier, metadata=metadata, created_at=created_at
        )

        return user



import html
from dataclasses import dataclass
from importlib.metadata import version
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional

import chevron
from typing_extensions import TypedDict, deprecated

if TYPE_CHECKING:
    from literalai.api import LiteralAPI

from literalai.my_types import Utils
from literalai.observability.generation import GenerationMessage, GenerationType


class ProviderSettings(TypedDict, total=False):
    provider: str
    model: str
    frequency_penalty: float
    max_tokens: int
    presence_penalty: float
    stop: Optional[List[str]]
    temperature: float
    top_p: float


class PromptVariable(TypedDict, total=False):
    name: str
    language: Literal["json", "plaintext"]


class LiteralMessageDict(dict):
    def __init__(self, prompt_id: str, variables: Dict, *args, **kwargs):
        super().__init__(*args, **kwargs)  # Initialize as a regular dict
        if "uuid" in self:
            uuid = self.pop("uuid")
            self.__literal_prompt__ = {
                "uuid": uuid,
                "prompt_id": prompt_id,
                "variables": variables,
            }


class PromptDict(TypedDict, total=False):
    id: str
    lineage: Dict
    createdAt: str
    updatedAt: str
    type: "GenerationType"
    name: str
    version: int
    versionDesc: Optional[str]
    templateMessages: List["GenerationMessage"]
    tools: Optional[List[Dict]]
    provider: str
    settings: ProviderSettings
    variables: List[PromptVariable]
    variablesDefaultValues: Optional[Dict[str, Any]]


@dataclass(repr=False)
class Prompt(Utils):
    api: "LiteralAPI"
    id: str
    created_at: str
    updated_at: str
    type: "GenerationType"
    name: str
    version: int
    version_desc: Optional[str]
    template_messages: List["GenerationMessage"]
    tools: Optional[List[Dict]]
    provider: str
    settings: ProviderSettings
    variables: List[PromptVariable]
    variables_default_values: Optional[Dict[str, Any]]

    def to_dict(self) -> PromptDict:
        # Convert the Prompt instance to a dictionary matching the PromptDict structure
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "updatedAt": self.updated_at,
            "type": self.type,
            "name": self.name,
            "version": self.version,
            "versionDesc": self.version_desc,
            "templateMessages": self.template_messages,  # Assuming this is a list of dicts or similar serializable objects
            "tools": self.tools,
            "provider": self.provider,
            "settings": self.settings,
            "variables": self.variables,
            "variablesDefaultValues": self.variables_default_values,
        }

    @classmethod
    def from_dict(cls, api: "LiteralAPI", prompt_dict: PromptDict) -> "Prompt":
        # Create a Prompt instance from a dictionary (PromptDict)
        settings = prompt_dict.get("settings") or {}
        provider = settings.pop("provider", "")

        return cls(
            api=api,
            id=prompt_dict.get("id", ""),
            name=prompt_dict.get("lineage", {}).get("name", ""),
            version=prompt_dict.get("version", 0),
            created_at=prompt_dict.get("createdAt", ""),
            updated_at=prompt_dict.get("updatedAt", ""),
            type=prompt_dict.get("type", GenerationType.CHAT),
            version_desc=prompt_dict.get("versionDesc"),
            template_messages=prompt_dict.get("templateMessages", []),
            tools=prompt_dict.get("tools", []),
            provider=provider,
            settings=settings,
            variables=prompt_dict.get("variables", []),
            variables_default_values=prompt_dict.get("variablesDefaultValues"),
        )

    def format_messages(self, **kwargs: Any) -> List[Any]:
        """
        Formats the prompt's template messages with the given variables.
        Variables may be passed as a dictionary or as keyword arguments.
        Keyword arguments take precedence over variables passed as a dictionary.

        Args:
            variables (Optional[Dict[str, Any]]): Optional variables to resolve in the template messages.


        Returns:
            List[Any]: List of formatted chat completion messages.
        """
        variables_with_defaults = {
            **(self.variables_default_values or {}),
            **(kwargs or {}),
        }
        formatted_messages = []

        for message in self.template_messages:
            formatted_message = LiteralMessageDict(
                self.id, variables_with_defaults, message.copy()
            )
            if isinstance(formatted_message["content"], str):
                formatted_message["content"] = html.unescape(
                    chevron.render(message["content"], variables_with_defaults)
                )
            else:
                for content in formatted_message["content"]:
                    if content["type"] == "text":
                        content["text"] = html.unescape(
                            chevron.render(content["text"], variables_with_defaults)
                        )

            formatted_messages.append(formatted_message)

        return formatted_messages

    @deprecated('Please use "format_messages" instead')
    def format(self, variables: Optional[Dict[str, Any]] = None) -> List[Any]:
        return self.format_messages(**(variables or {}))

    def to_langchain_chat_prompt_template(self):
        try:
            version("langchain")
        except Exception:
            raise Exception(
                "Please install langchain to use the langchain callback. "
                "You can install it with `pip install langchain`"
            )

        from langchain_core.messages import (
            AIMessage,
            BaseMessage,
            HumanMessage,
            SystemMessage,
        )
        from langchain_core.prompts import (
            AIMessagePromptTemplate,
            ChatPromptTemplate,
            HumanMessagePromptTemplate,
            SystemMessagePromptTemplate,
        )

        class CustomChatPromptTemplate(ChatPromptTemplate):
            orig_messages: Optional[List[GenerationMessage]]
            default_vars: Optional[Dict] = None
            prompt_id: Optional[str]

            def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
                variables_with_defaults = {
                    **(self.default_vars or {}),
                    **(kwargs or {}),
                }

                rendered_messages: List[BaseMessage] = []

                for index, message in enumerate(self.messages):
                    template = message.prompt.template  # type: ignore
                    content = html.unescape(
                        chevron.render(template, variables_with_defaults)
                    )
                    additonal_kwargs = {}
                    if self.orig_messages and index < len(self.orig_messages):
                        additonal_kwargs = {
                            "uuid": (
                                self.orig_messages[index].get("uuid")
                                if self.orig_messages
                                else None
                            ),
                            "prompt_id": self.prompt_id,
                            "variables": variables_with_defaults,
                        }

                    if isinstance(message, HumanMessagePromptTemplate):
                        rendered_messages.append(
                            HumanMessage(
                                content=content, additional_kwargs=additonal_kwargs
                            )
                        )
                    if isinstance(message, AIMessagePromptTemplate):
                        rendered_messages.append(
                            AIMessage(
                                content=content, additional_kwargs=additonal_kwargs
                            )
                        )
                    if isinstance(message, SystemMessagePromptTemplate):
                        rendered_messages.append(
                            SystemMessage(
                                content=content, additional_kwargs=additonal_kwargs
                            )
                        )

                return rendered_messages

        lc_messages = [(m["role"], m["content"]) for m in self.template_messages]

        chat_template = CustomChatPromptTemplate.from_messages(lc_messages)
        chat_template.input_variables = ["agent_scratchpad"]
        chat_template.default_vars = self.variables_default_values
        chat_template.orig_messages = self.template_messages
        chat_template.prompt_id = self.id

        return chat_template



import uuid
from typing import TYPE_CHECKING, Dict, List, Optional

if TYPE_CHECKING:
    from literalai.event_processor import EventProcessor

from literalai.context import active_steps_var, active_thread_var, active_root_run_var
from literalai.helper import utc_now
from literalai.my_types import Utils
from literalai.observability.step import MessageStepType, StepDict, Score, Attachment


class Message(Utils):
    id: Optional[str] = None
    name: Optional[str] = None
    type: Optional[MessageStepType] = None
    metadata: Optional[Dict] = {}
    parent_id: Optional[str] = None
    timestamp: Optional[str] = None
    content: str
    thread_id: Optional[str] = None
    root_run_id: Optional[str] = None
    tags: Optional[List[str]] = None
    created_at: Optional[str] = None

    scores: List[Score] = []
    attachments: List[Attachment] = []

    def __init__(
        self,
        content: str,
        id: Optional[str] = None,
        type: Optional[MessageStepType] = "assistant_message",
        name: Optional[str] = None,
        thread_id: Optional[str] = None,
        parent_id: Optional[str] = None,
        scores: List[Score] = [],
        attachments: List[Attachment] = [],
        metadata: Optional[Dict] = {},
        timestamp: Optional[str] = None,
        tags: Optional[List[str]] = [],
        processor: Optional["EventProcessor"] = None,
        root_run_id: Optional[str] = None,
    ):
        from time import sleep

        sleep(0.001)

        self.id = id or str(uuid.uuid4())
        if not timestamp:
            self.timestamp = utc_now()
        else:
            self.timestamp = timestamp
        self.name = name
        self.type = type
        self.content = content
        self.scores = scores
        self.attachments = attachments
        self.metadata = metadata
        self.tags = tags

        self.processor = processor

        # priority for thread_id: thread_id > parent_step.thread_id > active_thread
        self.thread_id = thread_id

        # priority for root_run_id: root_run_id > parent_step.root_run_id > active_root_run
        self.root_run_id = root_run_id

        # priority for parent_id: parent_id > parent_step.id
        self.parent_id = parent_id

    def end(self):
        active_steps = active_steps_var.get()

        if len(active_steps) > 0:
            parent_step = active_steps[-1]
            if not self.parent_id:
                self.parent_id = parent_step.id
            if not self.thread_id:
                self.thread_id = parent_step.thread_id
            if not self.root_run_id:
                self.root_run_id = parent_step.root_run_id

        if not self.thread_id:
            if active_thread := active_thread_var.get():
                self.thread_id = active_thread.id

        if not self.root_run_id:
            if active_root_run := active_root_run_var.get():
                self.root_run_id = active_root_run.id

        if not self.thread_id and not self.parent_id:
            raise Exception(
                "Message must be initialized with a thread_id or a parent id."
            )

        if self.processor is None:
            raise Exception(
                "Message must be initialized with a processor to allow finalization."
            )
        self.processor.add_event(self.to_dict())

    def to_dict(self) -> "StepDict":
        # Create a correct step Dict from a message
        return {
            "id": self.id,
            "metadata": self.metadata,
            "parentId": self.parent_id,
            "startTime": self.timestamp,
            "endTime": self.timestamp,  # startTime = endTime in Message
            "type": self.type,
            "threadId": self.thread_id,
            "output": {
                "content": self.content
            },  # no input, output = content in Message
            "name": self.name,
            "tags": self.tags,
            "scores": [score.to_dict() for score in self.scores],
            "attachments": [attachment.to_dict() for attachment in self.attachments],
            "rootRunId": self.root_run_id,
        }

    @classmethod
    def from_dict(cls, message_dict: Dict) -> "Message":
        id = message_dict.get("id", None)
        type = message_dict.get("type", None)
        thread_id = message_dict.get("threadId", None)
        root_run_id = message_dict.get("rootRunId", None)

        metadata = message_dict.get("metadata", None)
        parent_id = message_dict.get("parentId", None)
        timestamp = message_dict.get("startTime", None)
        content = message_dict.get("output", {}).get("content", "")
        name = message_dict.get("name", None)
        scores = message_dict.get("scores", [])
        attachments = message_dict.get("attachments", [])

        message = cls(
            id=id,
            metadata=metadata,
            parent_id=parent_id,
            timestamp=timestamp,
            type=type,
            thread_id=thread_id,
            content=content,
            name=name,
            scores=scores,
            attachments=attachments,
            root_run_id=root_run_id,
        )

        message.created_at = message_dict.get("createdAt", None)

        return message



from typing import Any, Dict, Optional

from literalai.observability.filter import users_filters
from literalai.my_types import PaginatedResponse, User

from literalai.api import gql


def get_users_helper(
    first: Optional[int] = None,
    after: Optional[str] = None,
    before: Optional[str] = None,
    filters: Optional[users_filters] = None,
):
    variables: Dict[str, Any] = {}

    if first:
        variables["first"] = first
    if after:
        variables["after"] = after
    if before:
        variables["before"] = before
    if filters:
        variables["filters"] = filters

    def process_response(response):
        response = response["data"]["participants"]
        response["data"] = list(map(lambda x: x["node"], response["edges"]))
        return PaginatedResponse[User].from_dict(response, User)

    description = "get users"

    return gql.GET_PARTICIPANTS, description, variables, process_response


def create_user_helper(identifier: str, metadata: Optional[Dict] = None):
    variables = {"identifier": identifier, "metadata": metadata}

    def process_response(response):
        return User.from_dict(response["data"]["createParticipant"])

    description = "create user"

    return gql.CREATE_PARTICIPANT, description, variables, process_response


def update_user_helper(
    id: str, identifier: Optional[str] = None, metadata: Optional[Dict] = None
):
    variables = {"id": id, "identifier": identifier, "metadata": metadata}

    # remove None values to prevent the API from removing existing values
    variables = {k: v for k, v in variables.items() if v is not None}

    def process_response(response):
        return User.from_dict(response["data"]["updateParticipant"])

    description = "update user"

    return gql.UPDATE_PARTICIPANT, description, variables, process_response


def get_user_helper(id: Optional[str] = None, identifier: Optional[str] = None):
    if id is None and identifier is None:
        raise Exception("Either id or identifier must be provided")

    if id is not None and identifier is not None:
        raise Exception("Only one of id or identifier must be provided")

    variables = {"id": id, "identifier": identifier}

    def process_response(response):
        user = response["data"]["participant"]
        return User.from_dict(user) if user else None

    description = "get user"

    return gql.GET_PARTICIPANT, description, variables, process_response


def delete_user_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        return response["data"]["deleteParticipant"]["id"]

    description = "delete user"

    return gql.DELETE_PARTICIPANT, description, variables, process_response



from typing import List, Union

from literalai.helper import ensure_values_serializable
from literalai.observability.step import Step, StepDict

STEP_FIELDS = """
        id
        threadId
        rootRunId
        parentId
        startTime
        endTime
        createdAt
        type
        error
        input
        output
        metadata
        environment
        scores {
            id
            type
            name
            value
            comment
        }
        tags
        generation {
          prompt
          completion
          createdAt
          provider
          model
          variables
          messages
          messageCompletion
          tools
          settings
          stepId
          tokenCount              
          inputTokenCount         
          outputTokenCount        
          ttFirstToken          
          duration                
          tokenThroughputInSeconds
          error
          type
        }
        name
        attachments {
            id
            stepId
            metadata
            mime
            name
            objectKey
            url
        }"""

THREAD_FIELDS = (
    """
        id
        name
        metadata
        tags
        createdAt
        participant {
            id
            identifier
            metadata
        }
        steps {
"""
    + STEP_FIELDS
    + """
        }"""
)

SHALLOW_THREAD_FIELDS = """
        id
        name
        metadata
        tags
        createdAt
        participant {
            id
            identifier
            metadata
        }    
        
"""

GET_PARTICIPANTS = """query GetParticipants(
$after: ID,
$before: ID,
$cursorAnchor: DateTime,
$filters: [participantsInputType!],
$first: Int,
$last: Int,
$projectId: String,
) {
participants(
    after: $after,
    before: $before,
    cursorAnchor: $cursorAnchor,
    filters: $filters,
    first: $first,
    last: $last,
    projectId: $projectId,
    ) {
    pageInfo {
        startCursor
        endCursor
        hasNextPage
        hasPreviousPage
    }
    totalCount
    edges {
        cursor
        node {
            id
            createdAt
            lastEngaged
            threadCount
            tokenCount
            identifier
            metadata
        }
    }
}
}
"""

CREATE_PARTICIPANT = """mutation CreateUser($identifier: String!, $metadata: Json) {
createParticipant(identifier: $identifier, metadata: $metadata) {
    id
    identifier
    metadata
}
}
"""

UPDATE_PARTICIPANT = """mutation UpdateUser(
    $id: String!,
    $identifier: String,
    $metadata: Json,
) {
    updateParticipant(
        id: $id,
        identifier: $identifier,
        metadata: $metadata
    ) {
        id
        identifier
        metadata
    }
}
"""

GET_PARTICIPANT = """query GetUser($id: String, $identifier: String) {
participant(id: $id, identifier: $identifier) {
    id
    identifier
    metadata
    createdAt
}
}"""

DELETE_PARTICIPANT = """
mutation DeleteUser($id: String!) {
    deleteParticipant(id: $id) {
        id
    }
}
"""

GET_THREADS = (
    """
query GetThreads(
    $after: ID,
    $before: ID,
    $cursorAnchor: DateTime,
    $filters: [ThreadsInputType!],
    $orderBy: ThreadsOrderByInput,
    $first: Int,
    $last: Int,
    $projectId: String,
    $stepTypesToKeep: [StepType!],
    ) {
    threads(
        after: $after,
        before: $before,
        cursorAnchor: $cursorAnchor,
        filters: $filters,
        orderBy: $orderBy,
        first: $first,
        last: $last,
        projectId: $projectId,
        stepTypesToKeep: $stepTypesToKeep,
        ) {
        pageInfo {
            startCursor
            endCursor
            hasNextPage
            hasPreviousPage
        }
        totalCount
        edges {
            cursor
            node {
"""
    + THREAD_FIELDS
    + """
            }
        }
    }
}
"""
)

LIST_THREADS = """query listThreads(
$first: Int
$after: ID
$last: Int
$before: ID
$skip: Int
$projectId: String
$filters: [ThreadsInputType!]
$orderBy: ThreadsOrderByInput
$cursorAnchor: DateTime
) {
threads(
    first: $first
    after: $after
    last: $last
    before: $before
    skip: $skip
    projectId: $projectId
    filters: $filters
    orderBy: $orderBy
    cursorAnchor: $cursorAnchor
) {
    pageInfo {
    hasNextPage
    hasPreviousPage
    startCursor
    endCursor
    }
    totalCount
    edges {
    node {
        id
        createdAt
        tokenCount
        name
        metadata
        duration
        tags
        participant {
        identifier
        id
        }
    }
    }
}
}"""

CREATE_THREAD = (
    """
mutation CreateThread(
    $name: String,
    $metadata: Json,
    $participantId: String,
    $tags: [String!],
) {
    createThread(
        name: $name
        metadata: $metadata
        participantId: $participantId
        tags: $tags
    ) {
"""
    + SHALLOW_THREAD_FIELDS
    + """
    }
}
"""
)

UPSERT_THREAD = (
    """
mutation UpsertThread(
    $id: String!,
    $name: String,
    $metadata: Json,
    $participantId: String,
    $tags: [String!],
) {
    upsertThread(
        id: $id
        name: $name
        metadata: $metadata
        participantId: $participantId
        tags: $tags
    ) {
"""
    + SHALLOW_THREAD_FIELDS
    + """
    }
}
"""
)

UPDATE_THREAD = (
    """
mutation UpdateThread(
    $id: String!,
    $name: String,
    $metadata: Json,
    $participantId: String,
    $tags: [String!],
) {
    updateThread(
        id: $id
        name: $name
        metadata: $metadata
        participantId: $participantId
        tags: $tags
    ) {
"""
    + SHALLOW_THREAD_FIELDS
    + """
    }
}
"""
)

GET_THREAD = (
    """
query GetThread($id: String!) {
    threadDetail(id: $id) {
"""
    + THREAD_FIELDS
    + """
    }
}
"""
)

DELETE_THREAD = """
mutation DeleteThread($thread_id: String!) {
    deleteThread(id: $thread_id) {
        id
    }
}
"""

GET_SCORES = """
query GetScores(
    $after: ID,
    $before: ID,
    $cursorAnchor: DateTime,
    $filters: [scoresInputType!],
    $orderBy: ScoresOrderByInput,
    $first: Int,
    $last: Int,
    $projectId: String,
    ) {
    scores(
        after: $after,
        before: $before,
        cursorAnchor: $cursorAnchor,
        filters: $filters,
        orderBy: $orderBy,
        first: $first,
        last: $last,
        projectId: $projectId,
        ) {
        pageInfo {
            startCursor
            endCursor
            hasNextPage
            hasPreviousPage
        }
        totalCount
        edges {
            cursor
            node {
                comment
                createdAt
                id
                projectId
                stepId
                datasetExperimentItemId
                type
                updatedAt
                name
                value
                tags
                step {
                    thread {
                    id
                    participant {
                        identifier
                            }
                        }
                    }
                }
            }
        }
    }
"""

CREATE_SCORE = """
mutation CreateScore(
    $name: String!,
    $type: ScoreType!,
    $value: Float!,
    $stepId: String,
    $datasetExperimentItemId: String,
    $comment: String,
    $tags: [String!],

) {
    createScore(
        name: $name,
        type: $type,
        value: $value,
        stepId: $stepId,
        datasetExperimentItemId: $datasetExperimentItemId,
        comment: $comment,
        tags: $tags,
    ) {
        id
        name,
        type,
        value,
        stepId,
        datasetExperimentItemId,
        comment,
        tags,
    }
}
"""

UPDATE_SCORE = """
mutation UpdateScore(
    $id: String!,
    $comment: String,
    $value: Float!,
) {
    updateScore(
        id: $id,
        comment: $comment,
        value: $value,
    ) {
        id
        name,
        type,
        value,
        stepId,
        datasetExperimentItemId,
        comment
    }
}
"""

DELETE_SCORE = """
mutation DeleteScore($id: String!) {
    deleteScore(id: $id) {
        id
    }
}
"""

CREATE_ATTACHMENT = """
mutation CreateAttachment(
    $metadata: Json,
    $mime: String,
    $name: String!,
    $objectKey: String,
    $stepId: String!,
    $url: String,
) {
    createAttachment(
        metadata: $metadata,
        mime: $mime,
        name: $name,
        objectKey: $objectKey,
        stepId: $stepId,
        url: $url,
    ) {
        id
        threadId
        stepId
        metadata
        mime
        name
        objectKey
        url
    }
}
"""

UPDATE_ATTACHMENT = """
mutation UpdateAttachment(
    $id: String!,
    $metadata: Json,
    $mime: String,
    $name: String,
    $objectKey: String,
    $projectId: String,
    $url: String,
) {
    updateAttachment(
        id: $id,
        metadata: $metadata,
        mime: $mime,
        name: $name,
        objectKey: $objectKey,
        projectId: $projectId,
        url: $url,
    ) {
        id
        threadId
        stepId
        metadata
        mime
        name
        objectKey
        url
    }
}
"""

GET_ATTACHMENT = """
query GetAttachment($id: String!) {
    attachment(id: $id) {
        id
        threadId
        stepId
        metadata
        mime
        name
        objectKey
        url
    }
}
"""

DELETE_ATTACHMENT = """
mutation DeleteAttachment($id: String!) {
    deleteAttachment(id: $id) {
        id
    }
}
"""

CREATE_STEP = (
    """
mutation CreateStep(
    $threadId: String,
    $rootRunId: String,
    $type: StepType,
    $startTime: DateTime,
    $endTime: DateTime,
    $input: Json,
    $output: Json,
    $metadata: Json,
    $parentId: String,
    $name: String,
    $tags: [String!],

) {
    createStep(
        threadId: $threadId,
        rootRunId: $rootRunId,
        type: $type,
        startTime: $startTime,
        endTime: $endTime,
        input: $input,
        output: $output,
        metadata: $metadata,
        parentId: $parentId,
        name: $name,
        tags: $tags,
    ) {
"""
    + STEP_FIELDS
    + """
    }
}
"""
)

UPDATE_STEP = (
    """
mutation UpdateStep(
    $id: String!,
    $type: StepType,
    $input: Json,
    $output: Json,
    $metadata: Json,
    $name: String,
    $startTime: DateTime,
    $endTime: DateTime,
    $parentId: String,
    $tags: [String!],
) {
    updateStep(
        id: $id,
        type: $type,
        startTime: $startTime,
        endTime: $endTime,
        input: $input,
        output: $output,
        metadata: $metadata,
        name: $name,
        tags: $tags,
        parentId: $parentId,
    ) {    
"""
    + STEP_FIELDS
    + """
    }
}
"""
)

GET_STEPS = (
    """
query GetSteps(
    $after: ID,
    $before: ID,
    $cursorAnchor: DateTime,
    $filters: [stepsInputType!],
    $orderBy: StepsOrderByInput,
    $first: Int,
    $last: Int,
    $projectId: String,
    ) {
    steps(
        after: $after,
        before: $before,
        cursorAnchor: $cursorAnchor,
        filters: $filters,
        orderBy: $orderBy,
        first: $first,
        last: $last,
        projectId: $projectId,
        ) {
        pageInfo {
            startCursor
            endCursor
            hasNextPage
            hasPreviousPage
        }
        totalCount
        edges {
            cursor
            node {
"""
    + STEP_FIELDS
    + """
            }
        }
    }
}
"""
)

GET_STEP = (
    """
query GetStep($id: String!) {
    step(id: $id) {"""
    + STEP_FIELDS
    + """
    }
}
"""
)

DELETE_STEP = """
mutation DeleteStep($id: String!) {
    deleteStep(id: $id) {
        id
    }
}
"""

GET_GENERATIONS = """
query GetGenerations(
    $after: ID,
    $before: ID,
    $cursorAnchor: DateTime,
    $filters: [generationsInputType!],
    $orderBy: GenerationsOrderByInput,
    $first: Int,
    $last: Int,
    $projectId: String,
    ) {
    generations(
        after: $after,
        before: $before,
        cursorAnchor: $cursorAnchor,
        filters: $filters,
        orderBy: $orderBy,
        first: $first,
        last: $last,
        projectId: $projectId,
        ) {
        pageInfo {
            startCursor
            endCursor
            hasNextPage
            hasPreviousPage
        }
        totalCount
        edges {
            cursor
            node {
                id
                projectId
                prompt
                completion
                createdAt
                provider
                model
                variables
                messages
                messageCompletion
                tools
                settings
                stepId
                tokenCount
                duration
                inputTokenCount
                outputTokenCount
                ttFirstToken
                duration
                tokenThroughputInSeconds
                error
                type
                tags
                metadata
                step {
                    threadId
                    thread {
                    participant {
                        identifier
                            }
                        }
                    }
                }
            }
        }
    }
"""

CREATE_GENERATION = """
mutation CreateGeneration($generation: GenerationPayloadInput!) {
    createGeneration(generation: $generation) {
        id,
        type
    }
}
"""

CREATE_DATASET = """
mutation createDataset(
    $name: String!
    $description: String
    $metadata: Json
    $type: DatasetType
) {
    createDataset(
        name: $name
        description: $description
        metadata: $metadata
        type: $type
    ) {
        id
        createdAt
        name
        description
        metadata
        type
    }
}
"""

UPDATE_DATASET = """
mutation UpdateDataset(
    $id: String!
    $name: String
    $description: String
    $metadata: Json
) {
    updateDataset(
        id: $id
        name: $name
        description: $description
        metadata: $metadata
    ) {
        id
        createdAt
        name
        description
        metadata
        type
    }
}
"""

DELETE_DATASET = """
    mutation DeleteDataset(
        $id: String!
    ) {
        deleteDataset(
            id: $id
        ) {
            id
            createdAt
            name
            description
            metadata
            type
        }
    }
"""

CREATE_EXPERIMENT = """
    mutation CreateDatasetExperiment(
        $name: String! 
        $datasetId: String
        $promptId: String
        $params: Json
    ) {
        createDatasetExperiment(
            name: $name
            datasetId: $datasetId
            promptId: $promptId
            params: $params
        ) {
          id
          name
          datasetId
          params
        }
    }
"""

CREATE_EXPERIMENT_ITEM = """
    mutation CreateDatasetExperimentItem(
        $datasetExperimentId: String!
        $experimentRunId: String
        $datasetItemId: String
        $input: Json
        $output: Json
    ) {
        createDatasetExperimentItem(
            datasetExperimentId: $datasetExperimentId
            datasetItemId: $datasetItemId
            experimentRunId: $experimentRunId
            input: $input
            output: $output
        ) {
          id
          input
          output
          datasetExperimentId
          experimentRunId
          datasetItemId
        }
      }
"""

CREATE_DATASET_ITEM = """
mutation CreateDatasetItem(
    $datasetId: String!
    $input: Json!
    $expectedOutput: Json
    $metadata: Json
    $generationId: String
) {
    createDatasetItem(
        datasetId: $datasetId
        input: $input
        expectedOutput: $expectedOutput
        metadata: $metadata
        generationId: $generationId
    ) {
        id
        createdAt
        datasetId
        metadata
        input
        expectedOutput
        intermediarySteps
        stepId
    }
}
"""

GET_DATASET_ITEM = """
query GetDataItem($id: String!) {
    datasetItem(id: $id) {
        id
        createdAt
        datasetId
        metadata
        input
        expectedOutput
        intermediarySteps
        stepId
    }
}
"""

DELETE_DATASET_ITEM = """
mutation DeleteDatasetItem($id: String!) {
    deleteDatasetItem(id: $id) {
        id
        createdAt
        datasetId
        metadata
        input
        expectedOutput
        intermediarySteps
        stepId
    }
}
"""

ADD_STEP_TO_DATASET = """
mutation AddStepToDataset(
    $datasetId: String!
    $stepId: String!
    $metadata: Json
) {
    addStepToDataset(
        datasetId: $datasetId
        stepId: $stepId
        metadata: $metadata
    ) {
        id
        createdAt
        datasetId
        metadata
        input
        expectedOutput
        intermediarySteps
        stepId
    }
}
"""

ADD_GENERATION_TO_DATASET = """
mutation AddGenerationToDataset(
    $datasetId: String!
    $generationId: String!
    $metadata: Json
) {
    addGenerationToDataset(
        datasetId: $datasetId
        generationId: $generationId
        metadata: $metadata
    ) {
        id
        createdAt
        datasetId
        metadata
        input
        expectedOutput
        intermediarySteps
        stepId
    }
}
"""

CREATE_PROMPT_LINEAGE = """mutation createPromptLineage(
    $name: String!
    $description: String
  ) {
    createPromptLineage(
      name: $name
      description: $description
    ) {
      id
      name
    }
  }"""

CREATE_PROMPT_VERSION = """mutation createPromptVersion(
    $lineageId: String!
    $versionDesc: String
    $templateMessages: Json
    $tools: Json
    $settings: Json
    $variables: Json
    $variablesDefaultValues: Json
  ) {
    createPromptVersion(
      lineageId: $lineageId
      versionDesc: $versionDesc
      templateMessages: $templateMessages
      tools: $tools
      settings: $settings
      variables: $variables
      variablesDefaultValues: $variablesDefaultValues
    ) {
      id
      lineage {
        name
      }
      version
      createdAt
      tools
      settings
      templateMessages
    }
  }"""

GET_PROMPT_VERSION = """
query GetPrompt($id: String, $name: String, $version: Int) {
    promptVersion(id: $id, name: $name, version: $version) {
        id
        createdAt
        updatedAt
        type
        templateMessages
        tools
        settings
        variables
        variablesDefaultValues
        version
        lineage {
            name
        }
    }
}
"""

GET_PROMPT_AB_TESTING = """query getPromptLineageRollout($projectId: String, $lineageName: String!) {
    promptLineageRollout(projectId: $projectId, lineageName: $lineageName) {
      pageInfo {
        startCursor
        endCursor
      }
      edges {
        node {
          version
          rollout
        }
      }
    }
  }
"""

UPDATE_PROMPT_AB_TESTING = """mutation updatePromptLineageRollout(
    $projectId: String
    $name: String!
    $rollouts: [PromptVersionRolloutInput!]!
  ) {
    updatePromptLineageRollout(
      projectId: $projectId
      name: $name
      rollouts: $rollouts
    ) {
      ok
      message
      errorCode
    }
  }"""


def serialize_step(event, id):
    result = {}

    for key, value in event.items():
        # Only keep the keys that are not None to avoid overriding existing values
        if value is not None:
            result[f"{key}_{id}"] = value

    return result


def steps_variables_builder(steps: List[Union["StepDict", "Step"]]):
    variables = {}
    for i in range(len(steps)):
        step = steps[i]
        if isinstance(step, Step):
            if step.input:
                step.input = ensure_values_serializable(step.input)
            if step.output:
                step.output = ensure_values_serializable(step.output)
            variables.update(serialize_step(step.to_dict(), i))
        else:
            if step.get("input"):
                step["input"] = ensure_values_serializable(step["input"])
            if step.get("output"):
                step["output"] = ensure_values_serializable(step["output"])
            variables.update(serialize_step(step, i))
    return variables


def steps_query_variables_builder(steps):
    generated = ""
    for id in range(len(steps)):
        generated += f"""$id_{id}: String!
        $threadId_{id}: String
        $rootRunId_{id}: String
        $type_{id}: StepType
        $startTime_{id}: DateTime
        $endTime_{id}: DateTime
        $error_{id}: String
        $input_{id}: Json
        $output_{id}: Json
        $metadata_{id}: Json
        $parentId_{id}: String
        $name_{id}: String
        $tags_{id}: [String!]
        $generation_{id}: GenerationPayloadInput
        $scores_{id}: [ScorePayloadInput!]
        $attachments_{id}: [AttachmentPayloadInput!]
        """
    return generated


def steps_ingest_steps_builder(steps):
    generated = ""
    for id in range(len(steps)):
        generated += f"""
      step{id}: ingestStep(
        id: $id_{id}
        threadId: $threadId_{id}
        rootRunId: $rootRunId_{id}
        startTime: $startTime_{id}
        endTime: $endTime_{id}
        type: $type_{id}
        error: $error_{id}
        input: $input_{id}
        output: $output_{id}
        metadata: $metadata_{id}
        parentId: $parentId_{id}
        name: $name_{id}
        tags: $tags_{id}
        generation: $generation_{id}
        scores: $scores_{id}
        attachments: $attachments_{id}
      ) {{
        ok
        message
      }}
"""
    return generated


def steps_query_builder(steps):
    return f"""
    mutation AddStep({steps_query_variables_builder(steps)}) {{
      {steps_ingest_steps_builder(steps)}
    }}
    """



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, cast

from literalai.my_types import Utils

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from literalai.api import LiteralAPI

from literalai.evaluation.dataset_experiment import DatasetExperiment
from literalai.evaluation.dataset_item import DatasetItem, DatasetItemDict

DatasetType = Literal["key_value", "generation"]


class DatasetDict(TypedDict, total=False):
    id: str
    createdAt: str
    metadata: Dict
    name: Optional[str]
    description: Optional[str]
    items: Optional[List[DatasetItemDict]]
    type: DatasetType


@dataclass(repr=False)
class Dataset(Utils):
    api: "LiteralAPI"
    id: str
    created_at: str
    metadata: Dict
    name: Optional[str] = None
    description: Optional[str] = None
    items: List[DatasetItem] = field(default_factory=lambda: [])
    type: DatasetType = "key_value"

    def to_dict(self):
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "metadata": self.metadata,
            "name": self.name,
            "description": self.description,
            "items": [item.to_dict() for item in self.items],
            "type": self.type,
        }

    @classmethod
    def from_dict(cls, api: "LiteralAPI", dataset: DatasetDict) -> "Dataset":
        items = dataset.get("items", [])
        if not isinstance(items, list):
            raise Exception("Dataset items should be an array")

        return cls(
            api=api,
            id=dataset.get("id", ""),
            created_at=dataset.get("createdAt", ""),
            metadata=dataset.get("metadata", {}),
            name=dataset.get("name"),
            description=dataset.get("description"),
            items=[DatasetItem.from_dict(item) for item in items],
            type=dataset.get("type", cast(DatasetType, "key_value")),
        )

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        metadata: Optional[Dict] = None,
    ):
        updated_dataset = self.api.update_dataset(
            self.id, name=name, description=description, metadata=metadata
        )
        self.name = updated_dataset.name
        self.description = updated_dataset.description
        self.metadata = updated_dataset.metadata

    def delete(self):
        self.api.delete_dataset(self.id)

    def create_item(
        self,
        input: Dict,
        expected_output: Optional[Dict] = None,
        metadata: Optional[Dict] = None,
    ) -> DatasetItem:
        """
        Create a new dataset item and add it to this dataset.
        :param input: The input data for the dataset item.
        :param expected_output: The output data for the dataset item (optional).
        :param metadata: Metadata for the dataset item (optional).
        :return: The created DatasetItem instance.
        """
        dataset_item = self.api.create_dataset_item(
            self.id, input, expected_output, metadata
        )
        if self.items is None:
            self.items = []
        self.items.append(dataset_item)
        return dataset_item

    def create_experiment(
        self, name: str, prompt_id: Optional[str] = None, params: Optional[Dict] = None
    ) -> DatasetExperiment:
        """
        Creates a new dataset experiment based on this dataset.
        :param name: The name of the experiment .
        :param prompt_id: The Prompt ID used on LLM calls (optional).
        :param params: The params used on the experiment.
        :return: The created DatasetExperiment instance.
        """
        experiment = self.api.create_experiment(
            name=name, dataset_id=self.id, prompt_id=prompt_id, params=params
        )
        return experiment

    def delete_item(self, item_id: str):
        """
        Delete a dataset item from this dataset.
        :param api: An instance of the DatasetAPI to make the call.
        :param item_id: The ID of the dataset item to delete.
        """
        self.api.delete_dataset_item(item_id)
        if self.items is not None:
            self.items = [item for item in self.items if item.id != item_id]

    def add_step(self, step_id: str, metadata: Optional[Dict] = None) -> DatasetItem:
        """
        Create a new dataset item based on a step and add it to this dataset.
        :param step_id: The id of the step to add to the dataset.
        :param metadata: Metadata for the dataset item (optional).
        :return: The created DatasetItem instance.
        """
        if self.type == "generation":
            raise ValueError("Cannot add a step to a generation dataset")

        dataset_item = self.api.add_step_to_dataset(self.id, step_id, metadata)
        if self.items is None:
            self.items = []
        self.items.append(dataset_item)
        return dataset_item

    def add_generation(
        self, generation_id: str, metadata: Optional[Dict] = None
    ) -> DatasetItem:
        """
        Create a new dataset item based on a generation and add it to this dataset.
        :param generation_id: The id of the generation to add to the dataset.
        :param metadata: Metadata for the dataset item (optional).
        :return: The created DatasetItem instance.
        """
        dataset_item = self.api.add_generation_to_dataset(
            self.id, generation_id, metadata
        )
        if self.items is None:
            self.items = []
        self.items.append(dataset_item)
        return dataset_item



from dataclasses import dataclass
from typing import Dict, List, Optional, TypedDict

from literalai.my_types import Utils


class DatasetItemDict(TypedDict, total=False):
    id: str
    createdAt: str
    datasetId: str
    metadata: Dict
    input: Dict
    expectedOutput: Optional[Dict]
    intermediarySteps: Optional[List[Dict]]


@dataclass(repr=False)
class DatasetItem(Utils):
    id: str
    created_at: str
    dataset_id: str
    metadata: Dict
    input: Dict
    expected_output: Optional[Dict] = None
    intermediary_steps: Optional[List[Dict]] = None

    def to_dict(self):
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "datasetId": self.dataset_id,
            "metadata": self.metadata,
            "input": self.input,
            "expectedOutput": self.expected_output,
            "intermediarySteps": self.intermediary_steps,
        }

    @classmethod
    def from_dict(cls, dataset_item: DatasetItemDict) -> "DatasetItem":
        return cls(
            id=dataset_item.get("id", ""),
            created_at=dataset_item.get("createdAt", ""),
            dataset_id=dataset_item.get("datasetId", ""),
            metadata=dataset_item.get("metadata", {}),
            input=dataset_item.get("input", {}),
            expected_output=dataset_item.get("expectedOutput"),
            intermediary_steps=dataset_item.get("intermediarySteps"),
        )



from typing_extensions import TypedDict

from typing import Any, Generic, List, Literal, Optional, TypeVar, Union

Field = TypeVar("Field")
Operators = TypeVar("Operators")
Value = TypeVar("Value")

BOOLEAN_OPERATORS = Literal["is", "nis"]
STRING_OPERATORS = Literal["eq", "neq", "ilike", "nilike"]
NUMBER_OPERATORS = Literal["eq", "neq", "gt", "gte", "lt", "lte"]
STRING_LIST_OPERATORS = Literal["in", "nin"]
DATETIME_OPERATORS = Literal["gte", "lte", "gt", "lt"]

OPERATORS = Union[
    BOOLEAN_OPERATORS,
    STRING_OPERATORS,
    NUMBER_OPERATORS,
    STRING_LIST_OPERATORS,
    DATETIME_OPERATORS,
]


class Filter(Generic[Field], TypedDict, total=False):
    field: Field
    operator: OPERATORS
    value: Any
    path: Optional[str]


class OrderBy(Generic[Field], TypedDict):
    column: Field
    direction: Literal["ASC", "DESC"]


threads_filterable_fields = Literal[
    "id",
    "createdAt",
    "name",
    "stepType",
    "stepName",
    "stepOutput",
    "metadata",
    "tokenCount",
    "tags",
    "participantId",
    "participantIdentifiers",
    "scoreValue",
    "duration",
]
threads_orderable_fields = Literal["createdAt", "tokenCount"]
threads_filters = List[Filter[threads_filterable_fields]]
threads_order_by = OrderBy[threads_orderable_fields]

steps_filterable_fields = Literal[
    "id",
    "name",
    "input",
    "output",
    "participantIdentifier",
    "startTime",
    "endTime",
    "metadata",
    "parentId",
    "threadId",
    "error",
    "tags",
]
steps_orderable_fields = Literal["createdAt"]
steps_filters = List[Filter[steps_filterable_fields]]
steps_order_by = OrderBy[steps_orderable_fields]

users_filterable_fields = Literal[
    "id",
    "createdAt",
    "identifier",
    "lastEngaged",
    "threadCount",
    "tokenCount",
    "metadata",
]
users_filters = List[Filter[users_filterable_fields]]

scores_filterable_fields = Literal[
    "id",
    "createdAt",
    "participant",
    "name",
    "tags",
    "value",
    "type",
    "comment",
]
scores_orderable_fields = Literal["createdAt"]
scores_filters = List[Filter[scores_filterable_fields]]
scores_order_by = OrderBy[scores_orderable_fields]

generation_filterable_fields = Literal[
    "id",
    "createdAt",
    "model",
    "duration",
    "promptLineage",
    "promptVersion",
    "tags",
    "score",
    "participant",
    "tokenCount",
    "error",
]
generation_orderable_fields = Literal[
    "createdAt",
    "tokenCount",
    "model",
    "provider",
    "participant",
    "duration",
]
generations_filters = List[Filter[generation_filterable_fields]]
generations_order_by = OrderBy[generation_orderable_fields]



import mimetypes
from typing import Dict, Optional, TypedDict, Union

from literalai.observability.step import Attachment

from literalai.api import gql


def create_attachment_helper(
    step_id: str,
    thread_id: Optional[str] = None,
    id: Optional[str] = None,
    metadata: Optional[Dict] = None,
    mime: Optional[str] = None,
    name: Optional[str] = None,
    object_key: Optional[str] = None,
    url: Optional[str] = None,
    content: Optional[Union[bytes, str]] = None,
    path: Optional[str] = None,
):
    if not content and not url and not path:
        raise Exception("Either content, path or attachment url must be provided")

    if content and path:
        raise Exception("Only one of content and path must be provided")

    if (content and url) or (path and url):
        raise Exception("Only one of content, path and attachment url must be provided")

    if path:
        # TODO: if attachment.mime is text, we could read as text?
        with open(path, "rb") as f:
            content = f.read()
        if not name:
            name = path.split("/")[-1]
        if not mime:
            mime, _ = mimetypes.guess_type(path)
            mime = mime or "application/octet-stream"

    if not name:
        raise Exception("Attachment name must be provided")

    variables = {
        "metadata": metadata,
        "mime": mime,
        "name": name,
        "objectKey": object_key,
        "stepId": step_id,
        "threadId": thread_id,
        "url": url,
        "id": id,
    }

    description = "create attachment"

    def process_response(response):
        return Attachment.from_dict(response["data"]["createAttachment"])

    return gql.CREATE_ATTACHMENT, description, variables, content, process_response


class AttachmentUpload(TypedDict, total=False):
    metadata: Optional[Dict]
    name: Optional[str]
    mime: Optional[str]
    objectKey: Optional[str]
    url: Optional[str]


def update_attachment_helper(id: str, update_params: AttachmentUpload):
    variables = {"id": id, **update_params}

    def process_response(response):
        return Attachment.from_dict(response["data"]["updateAttachment"])

    description = "update attachment"

    return gql.UPDATE_ATTACHMENT, description, variables, process_response


def get_attachment_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        attachment = response["data"]["attachment"]
        return Attachment.from_dict(attachment) if attachment else None

    description = "get attachment"

    return gql.GET_ATTACHMENT, description, variables, process_response


def delete_attachment_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        return response["data"]["deleteAttachment"]

    description = "delete attachment"

    return gql.DELETE_ATTACHMENT, description, variables, process_response



from typing import Any, Dict, List, Optional, Union

from literalai.observability.filter import steps_filters, steps_order_by
from literalai.my_types import PaginatedResponse
from literalai.observability.step import Step, StepDict, StepType

from literalai.api import gql


def create_step_helper(
    thread_id: Optional[str] = None,
    type: Optional[StepType] = "undefined",
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    input: Optional[Dict] = None,
    output: Optional[Dict] = None,
    metadata: Optional[Dict] = None,
    parent_id: Optional[str] = None,
    name: Optional[str] = None,
    tags: Optional[List[str]] = None,
    root_run_id: Optional[str] = None,
):
    variables = {
        "threadId": thread_id,
        "type": type,
        "startTime": start_time,
        "endTime": end_time,
        "input": input,
        "output": output,
        "metadata": metadata,
        "parentId": parent_id,
        "name": name,
        "tags": tags,
        "root_run_id": root_run_id,
    }

    def process_response(response):
        return Step.from_dict(response["data"]["createStep"])

    description = "create step"

    return gql.CREATE_STEP, description, variables, process_response


def update_step_helper(
    id: str,
    type: Optional[StepType] = None,
    input: Optional[str] = None,
    output: Optional[str] = None,
    metadata: Optional[Dict] = None,
    name: Optional[str] = None,
    tags: Optional[List[str]] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    parent_id: Optional[str] = None,
):
    variables = {
        "id": id,
        "type": type,
        "input": input,
        "output": output,
        "metadata": metadata,
        "name": name,
        "tags": tags,
        "startTime": start_time,
        "endTime": end_time,
        "parentId": parent_id,
    }

    def process_response(response):
        return Step.from_dict(response["data"]["updateStep"])

    description = "update step"

    return gql.UPDATE_STEP, description, variables, process_response


def get_steps_helper(
    first: Optional[int] = None,
    after: Optional[str] = None,
    before: Optional[str] = None,
    filters: Optional[steps_filters] = None,
    order_by: Optional[steps_order_by] = None,
):
    variables: Dict[str, Any] = {}

    if first:
        variables["first"] = first
    if after:
        variables["after"] = after
    if before:
        variables["before"] = before
    if filters:
        variables["filters"] = filters
    if order_by:
        variables["orderBy"] = order_by

    def process_response(response):
        processed_response = response["data"]["steps"]
        processed_response["data"] = [
            edge["node"] for edge in processed_response["edges"]
        ]
        return PaginatedResponse[Step].from_dict(processed_response, Step)

    description = "get steps"

    return gql.GET_STEPS, description, variables, process_response


def get_step_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        step = response["data"]["step"]
        return Step.from_dict(step) if step else None

    description = "get step"

    return gql.GET_STEP, description, variables, process_response


def delete_step_helper(id: str):
    variables = {"id": id}

    def process_response(response):
        return bool(response["data"]["deleteStep"])

    description = "delete step"

    return gql.DELETE_STEP, description, variables, process_response


def send_steps_helper(steps: List[Union[StepDict, "Step"]]):
    query = gql.steps_query_builder(steps)
    variables = gql.steps_variables_builder(steps)

    description = "send steps"

    def process_response(response: Dict):
        return response

    return query, description, variables, process_response



import time
from functools import wraps
from importlib import import_module
from typing import TYPE_CHECKING, Callable, Optional, TypedDict, Union

from literalai.context import active_steps_var

if TYPE_CHECKING:
    from literalai.observability.step import Step
    from literalai.observability.generation import CompletionGeneration, ChatGeneration


class BeforeContext(TypedDict):
    original_func: Callable
    generation: Optional[Union["ChatGeneration", "CompletionGeneration"]]
    step: Optional["Step"]
    start: float


class AfterContext(TypedDict):
    original_func: Callable
    generation: Optional[Union["ChatGeneration", "CompletionGeneration"]]
    step: Optional["Step"]
    start: float


def remove_literalai_args(kargs):
    '''Remove argument prefixed with "literalai_" from kwargs and return them in a separate dict'''
    largs = {}
    for key in list(kargs.keys()):
        if key.startswith("literalai_"):
            value = kargs.pop(key)
            largs[key] = value
    return largs

def restore_literalai_args(kargs, largs):
    '''Reverse the effect of remove_literalai_args by merging the literal arguments into kwargs'''
    for key in list(largs.keys()):
        kargs[key] = largs[key]


def sync_wrapper(before_func=None, after_func=None):
    def decorator(original_func):
        @wraps(original_func)
        def wrapped(*args, **kwargs):
            context = {"original_func": original_func}
            # If a before_func is provided, call it with the shared context.
            if before_func:
                before_func(context, *args, **kwargs)
            # Remove literal arguments before calling the original function
            literalai_args = remove_literalai_args(kwargs)
            context["start"] = time.time()
            try:
                result = original_func(*args, **kwargs)
            except Exception as e:
                active_steps = active_steps_var.get()
                if active_steps and len(active_steps) > 0:
                    current_step = active_steps[-1]
                    current_step.error = str(e)
                    current_step.end()
                    current_step.processor.flush()
                raise e
            # If an after_func is provided, call it with the result and the shared context.
            if after_func:
                restore_literalai_args(kwargs, literalai_args)
                result = after_func(result, context, *args, **kwargs)

            return result

        return wrapped

    return decorator


def async_wrapper(before_func=None, after_func=None):
    def decorator(original_func):
        @wraps(original_func)
        async def wrapped(*args, **kwargs):
            context = {"original_func": original_func}
            # If a before_func is provided, call it with the shared context.
            if before_func:
                await before_func(context, *args, **kwargs)

            # Remove literal arguments before calling the original function
            literalai_args = remove_literalai_args(kwargs)
            context["start"] = time.time()
            try:
                result = await original_func(*args, **kwargs)
            except Exception as e:
                active_steps = active_steps_var.get()
                if active_steps and len(active_steps) > 0:
                    current_step = active_steps[-1]
                    current_step.error = str(e)
                    current_step.end()
                    current_step.processor.flush()
                raise e

            # If an after_func is provided, call it with the result and the shared context.
            if after_func:
                restore_literalai_args(kwargs, literalai_args)
                result = await after_func(result, context, *args, **kwargs)

            return result

        return wrapped

    return decorator


def wrap_all(
    to_wrap: list,
    before_wrapper,
    after_wrapper,
    async_before_wrapper,
    async_after_wrapper,
):
    for patch in to_wrap:
        module = import_module(str(patch["module"]))
        target_object = getattr(module, str(patch["object"]))
        original_method = getattr(target_object, str(patch["method"]))

        if patch["async"]:
            wrapped_method = async_wrapper(
                before_func=async_before_wrapper(metadata=patch["metadata"]),
                after_func=async_after_wrapper(metadata=patch["metadata"]),
            )(original_method)
        else:
            wrapped_method = sync_wrapper(
                before_func=before_wrapper(metadata=patch["metadata"]),
                after_func=after_wrapper(metadata=patch["metadata"]),
            )(original_method)

        setattr(target_object, str(patch["method"]), wrapped_method)



import inspect
import uuid
from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional

from literalai.context import active_experiment_item_run_id_var
from literalai.environment import EnvContextManager
from literalai.observability.step import StepContextManager

if TYPE_CHECKING:
    from literalai.client import BaseLiteralClient


class ExperimentItemRunContextManager(EnvContextManager, StepContextManager):
    def __init__(
        self,
        client: "BaseLiteralClient",
    ):
        self.id = str(uuid.uuid4())
        EnvContextManager.__init__(self, client=client, env="experiment")
        StepContextManager.__init__(
            self, client=client, name="Experiment Run", type="run", id=self.id
        )

    def __call__(self, func):
        return experiment_item_run_decorator(
            self.client,
            func=func,
            ctx_manager=self,
        )

    async def __aenter__(self):
        active_experiment_item_run_id_var.set(self.id)
        await EnvContextManager.__aenter__(self)
        step = await StepContextManager.__aenter__(self)
        return step

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await StepContextManager.__aexit__(self, exc_type, exc_val, exc_tb)
        await self.client.event_processor.aflush()
        await EnvContextManager.__aexit__(self, exc_type, exc_val, exc_tb)
        active_experiment_item_run_id_var.set(None)

    def __enter__(self):
        active_experiment_item_run_id_var.set(self.id)
        EnvContextManager.__enter__(self)
        step = StepContextManager.__enter__(self)
        return step

    def __exit__(self, exc_type, exc_val, exc_tb):
        StepContextManager.__exit__(self, exc_type, exc_val, exc_tb)
        self.client.event_processor.flush()
        EnvContextManager.__exit__(self, exc_type, exc_val, exc_tb)
        active_experiment_item_run_id_var.set(None)


def experiment_item_run_decorator(
    client: "BaseLiteralClient",
    func: Callable,
    ctx_manager: Optional[ExperimentItemRunContextManager] = None,
    **decorator_kwargs,
):
    if not ctx_manager:
        ctx_manager = ExperimentItemRunContextManager(
            client=client,
            **decorator_kwargs,
        )

    # Handle async decorator
    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            with ctx_manager:
                result = await func(*args, **kwargs)
                return result

        return async_wrapper
    else:
        # Handle sync decorator
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            with ctx_manager:
                return func(*args, **kwargs)

        return sync_wrapper



from typing import TYPE_CHECKING, Dict, List, Optional, TypedDict

from literalai.observability.generation import GenerationMessage
from literalai.prompt_engineering.prompt import Prompt, ProviderSettings

if TYPE_CHECKING:
    from literalai.api import LiteralAPI

from literalai.api import gql


def create_prompt_lineage_helper(name: str, description: Optional[str] = None):
    variables = {"name": name, "description": description}

    def process_response(response):
        prompt = response["data"]["createPromptLineage"]
        return prompt

    description = "create prompt lineage"

    return gql.CREATE_PROMPT_LINEAGE, description, variables, process_response


def create_prompt_helper(
    api: "LiteralAPI",
    lineage_id: str,
    template_messages: List[GenerationMessage],
    settings: Optional[ProviderSettings] = None,
    tools: Optional[List[Dict]] = None,
):
    variables = {
        "lineageId": lineage_id,
        "templateMessages": template_messages,
        "settings": settings,
        "tools": tools,
    }

    def process_response(response):
        prompt = response["data"]["createPromptVersion"]
        return Prompt.from_dict(api, prompt) if prompt else None

    description = "create prompt version"

    return gql.CREATE_PROMPT_VERSION, description, variables, process_response


def get_prompt_helper(
    api: "LiteralAPI",
    id: Optional[str] = None,
    name: Optional[str] = None,
    version: Optional[int] = 0,
):
    variables = {"id": id, "name": name, "version": version}

    def process_response(response):
        prompt = response["data"]["promptVersion"]
        return Prompt.from_dict(api, prompt) if prompt else None

    description = "get prompt"

    return gql.GET_PROMPT_VERSION, description, variables, process_response


class PromptRollout(TypedDict):
    version: int
    rollout: int


def get_prompt_ab_testing_helper(
    name: Optional[str] = None,
):
    variables = {"lineageName": name}

    def process_response(response) -> List[PromptRollout]:
        response_data = response["data"]["promptLineageRollout"]
        return list(map(lambda x: x["node"], response_data["edges"]))

    description = "get prompt A/B testing"

    return gql.GET_PROMPT_AB_TESTING, description, variables, process_response


def update_prompt_ab_testing_helper(name: str, rollouts: List[PromptRollout]):
    variables = {"name": name, "rollouts": rollouts}

    def process_response(response) -> Dict:
        return response["data"]["updatePromptLineageRollout"]

    description = "update prompt A/B testing"

    return gql.UPDATE_PROMPT_AB_TESTING, description, variables, process_response
